#!/usr/bin/env python3
import os
import json
import subprocess
import shutil
import time
import paramiko
import warnings
from datetime import datetime
from argparse import ArgumentParser
from ftplib import FTP, error_perm, error_temp
from webdav3.client import Client as WebDavClient
from pod2gen import Podcast, Episode, Media, Category, Person

warnings.filterwarnings("ignore", category=UserWarning)

def log_message(message):
    current_time = datetime.now().strftime("[%d.%m.%Y, %H:%M:%S]")
    print(f"{current_time} {message}")

def load_config():
    dir_path = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(dir_path, 'config.json')
    with open(config_path, 'r') as file:
        return json.load(file)

def load_stations():
    dir_path = os.path.dirname(os.path.abspath(__file__))
    stations_path = os.path.join(dir_path, 'stations.json')
    with open(stations_path, 'r') as file:
        return json.load(file)

def normalize_filename(name):
    mapping = str.maketrans('ąćęłńóśźżĄĆĘŁŃÓŚŹŻ', 'acelnoszzACELNOSZZ')
    normalized_name = name.translate(mapping)
    normalized_name = ''.join(c for c in normalized_name if c.isalnum() or c in [' ', '-', '.'])
    return normalized_name

def group_destinations_by_params(config, destinations):
    """
    Grupuje destynacje według wymagań formatu, bitrate i liczby kanałów.
    """
    conversion_map = {}
    for destination in destinations:
        dest_config = config[destination]
        params = (dest_config['format'], dest_config['bitrate'], dest_config['channels'])
        if params not in conversion_map:
            conversion_map[params] = []
        conversion_map[params].append(destination)
    return conversion_map

def convert_audio_once(source_file, config, conversion_map):
    """
    Wykonuje konwersję audio tylko raz dla każdej unikalnej konfiguracji parametrów.
    Generuje pliki o nazwach w formacie:
    Test2 (27.12.24 14.49) [16kbps 1ch].mp3
    """
    converted_files = {}
    base_name, _ = os.path.splitext(os.path.basename(source_file))  # Usunięcie rozszerzenia .ts

    for params, destinations in conversion_map.items():
        target_format, target_bitrate, channels = params
        # Dodanie parametrów do istniejącej nazwy
        target_filename = f"{base_name} [{target_bitrate}kbps].{target_format}"
        target_path = os.path.join(os.path.dirname(source_file), target_filename)

        if os.path.exists(target_path):
            log_message(f"File already exists, skipping conversion: {target_path}")
        else:
            log_message(f"Converting file to {target_format.upper()} with {target_bitrate}kbps.")
            codec = get_codec_for_format(target_format)
            extra_params = get_extra_params_for_format(target_format)

            ffmpeg_command = [
                "ffmpeg",
                "-i", source_file,
                "-c:a", codec,
                "-b:a", f"{target_bitrate}k",
                "-ac", str(channels)
            ] + extra_params + ["-y", target_path]
            subprocess.run(ffmpeg_command, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
            log_message(f"Conversion completed: {target_path}")

        converted_files[tuple(params)] = target_path
    return converted_files


def get_codec_for_format(format):
    """
    Zwraca odpowiedni kodek dla danego formatu.
    """
    if format == "m4a":
        return "libfdk_aac"
    elif format == "mp3":
        return "libmp3lame"
    elif format == "opus":
        return "libopus"
    else:
        raise ValueError(f"Unsupported format: {format}")

def get_extra_params_for_format(format):
    """
    Zwraca dodatkowe parametry dla kodeka.
    """
    if format == "m4a":
        return ["-profile:a", "aac_he"]
    elif format == "opus":
        return ["-vbr", "on"]
    return []

def process_destinations(converted_files, config, station, show_name, conversion_map):
    """
    Wysyła przekonwertowane pliki do odpowiednich destynacji.
    """
    for params, file_path in converted_files.items():
        format, bitrate, channels = params
        log_message(f"Processing destinations for file: {file_path}")
        
        # Flaga kontrolująca, czy `local` jest jedną z destynacji
        keep_local_file = False

        for destination in conversion_map[params]:
            if destination == "ftp":
                upload_ftp(file_path, config, station, show_name)
                rss(config, f"{config['ftp']['path']}/{normalize_filename(station)}/{normalize_filename(show_name)}", station, show_name, 'ftp')
            elif destination == "sftp":
                upload_sftp(file_path, config, station, show_name)
                rss(config, f"{config['sftp']['path']}/{normalize_filename(station)}/{normalize_filename(show_name)}", station, show_name, 'sftp')
            elif destination == "nextcloud":
                upload_nextcloud(file_path, config, station, show_name)
            elif destination == "local":
                keep_local_file = True
                manage_local_space(os.path.dirname(file_path), config)
        
        # Usuwamy plik tylko wtedy, gdy nie ma go w lokalnej destynacji
        if not keep_local_file and os.path.exists(file_path):
            os.remove(file_path)
            log_message(f"Deleted local file after processing all destinations: {file_path}")

def manage_local_space(directory, config):
    max_files = config['local']['max_files']
    if max_files == 0:
        return

    files = [os.path.join(directory, f) for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
    audio_files = [f for f in files if f.endswith('.mp3') or f.endswith('.opus') or f.endswith('.m4a')]

    while len(audio_files) > max_files:
        oldest_file = min(audio_files, key=os.path.getmtime)
        os.remove(oldest_file)
        log_message(f"Deleted oldest local file: {oldest_file}")
        audio_files.remove(oldest_file)

def manage_ftp_space(ftp, folder_path, config):
    max_files = config['ftp']['max_files']
    if max_files == 0:
        return

    ftp.cwd(folder_path)
    files = ftp.nlst()
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]

    while len(audio_files) > max_files:
        oldest_file = min(audio_files, key=lambda x: ftp.sendcmd('MDTM ' + x))
        ftp.delete(oldest_file)
        log_message(f"Deleted oldest file on FTP: {oldest_file}")
        audio_files.remove(oldest_file)

def manage_sftp_space(sftp, folder_path, config):
    max_files = config['sftp']['max_files']
    if max_files == 0:
        return

    files = sftp.listdir(folder_path)
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]

    if len(audio_files) <= max_files:
        return

    files_info = [(file, sftp.stat(os.path.join(folder_path, file)).st_mtime) for file in audio_files]
    files_info.sort(key=lambda x: x[1])

    for file_info in files_info[:-max_files]:
        sftp.remove(os.path.join(folder_path, file_info[0]))
        log_message(f"Deleted oldest file on SFTP: {file_info[0]}")

def manage_nextcloud_space(client, folder_path, config):
    max_files = config['nextcloud']['max_files']
    if max_files == 0:
        return

    files = client.list(folder_path)
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]

    if len(audio_files) <= max_files:
        return

    files_info = [(file, client.info(os.path.join(folder_path, file)).get('modified')) for file in audio_files]
    files_info.sort(key=lambda x: x[1])

    for file_info in files_info[:-max_files]:
        client.clean(os.path.join(folder_path, file_info[0]))
        log_message(f"Deleted oldest file on Nextcloud: {file_info[0]}")

def upload_ftp(filename, config, station, show_name):
    attempts = 0
    success = False
    while not success and attempts < 5:
        try:
            target_config = config['ftp']
            with FTP(target_config['hostname']) as ftp:
                ftp.login(target_config['username'], target_config['password'])
                remote_dir = f"{target_config['path']}/{normalize_filename(station)}/{normalize_filename(show_name)}"

                parts = remote_dir.split('/')
                current_dir = ''
                for part in parts:
                    if part:
                        current_dir += f'/{part}'
                        try:
                            ftp.cwd(current_dir)
                        except error_perm as e:
                            if 'No such file or directory' in str(e):
                                ftp.mkd(current_dir)
                                ftp.cwd(current_dir)
                            else:
                                raise

                with open(filename, 'rb') as file:
                    ftp.storbinary('STOR ' + os.path.basename(filename), file)
                success = True
                manage_ftp_space(ftp, remote_dir, config)
        except (error_temp, BrokenPipeError) as e:
            attempts += 1
            log_message(f"Error: {e}. Retrying... ({attempts})")
        except Exception as e:
            log_message(f"Unexpected error: {e}")
            break

def upload_sftp(filename, config, station, show_name):
    attempts = 0
    success = False
    while not success and attempts < 5:
        try:
            target_config = config['sftp']
            transport = paramiko.Transport((target_config['hostname'].split(':')[0], int(target_config['hostname'].split(':')[1])))
            transport.connect(username=target_config['username'], password=target_config['password'])
            sftp = paramiko.SFTPClient.from_transport(transport)

            remote_dir = f"{target_config['path']}/{normalize_filename(station)}/{normalize_filename(show_name)}"
            try:
                sftp.chdir(remote_dir)
            except IOError:
                parts = remote_dir.split('/')
                current_dir = ''
                for part in parts:
                    if part:
                        current_dir += f'/{part}'
                        try:
                            sftp.chdir(current_dir)
                        except IOError:
                            sftp.mkdir(current_dir)
                            sftp.chdir(current_dir)

            sftp.put(filename, os.path.basename(filename))
            success = True
            log_message(f"Uploaded file to SFTP.")
            manage_sftp_space(sftp, remote_dir, config)
        except Exception as e:
            attempts += 1
            log_message(f"Error: {e}. Retrying... ({attempts})")
        finally:
            sftp.close()
            transport.close()

def upload_nextcloud(filename, config, station, show_name):
    options = {
        'webdav_hostname': config['nextcloud']['hostname'],
        'webdav_login': config['nextcloud']['username'],
        'webdav_password': config['nextcloud']['password']
    }
    client = WebDavClient(options)
    remote_dir = f"/public/{normalize_filename(station)}/{normalize_filename(show_name)}/"
    remote_path = remote_dir + os.path.basename(filename)

    ensure_remote_path(client, remote_path)
    client.upload_sync(remote_path=remote_path, local_path=filename)
    log_message(f"Uploaded file to NextCloud.")
    manage_nextcloud_space(client, remote_dir, config)

def ensure_remote_path(client, remote_path):
    parts = remote_path.strip('/').split('/')
    path = ''
    for part in parts[:-1]:
        path += f"/{part}"
        if not client.check(path):
            client.mkdir(path)

def rss(config, remote_dir, station, show_name, destination):
    """
    Generuje RSS dla podcastów na podstawie plików w zdalnej lokalizacji.
    """
    target_config = config[destination]
    web_base_url = target_config['url']

    # Tworzenie obiektu podcastu
    podcast = Podcast()
    podcast.name = f"{show_name} - {station}"
    podcast.authors = [Person(station)]
    podcast.email = "example@example.com"
    podcast.website = os.path.join(web_base_url, normalize_filename(station), normalize_filename(show_name)).replace(" ", "%20")
    podcast.description = f"Podcast {show_name} z {station}"
    podcast.language = "pl"
    podcast.category = Category("News", "Politics")
    podcast.explicit = False
    podcast.image = os.path.join(web_base_url, normalize_filename(station), normalize_filename(show_name), 'podcast.jpg').replace(" ", "%20")

    files_info = []

    # Pobieranie listy plików w zależności od destynacji
    if destination == 'ftp':
        with FTP(target_config['hostname']) as ftp:
            ftp.login(target_config['username'], target_config['password'])
            ftp.cwd(remote_dir)
            ftp.set_pasv(True)
            files = ftp.nlst()
            for file in files:
                if file.endswith('.mp3') or file.endswith('.m4a'):
                    size = ftp.size(file)
                    files_info.append({'name': file, 'size': size})
    elif destination == 'sftp':
        try:
            transport = paramiko.Transport((target_config['hostname'].split(':')[0], int(target_config['hostname'].split(':')[1])))
            transport.connect(username=target_config['username'], password=target_config['password'])
            sftp = paramiko.SFTPClient.from_transport(transport)
            sftp.chdir(remote_dir)
            for file_attr in sftp.listdir_attr():
                file_name = file_attr.filename
                if file_name.endswith('.mp3') or file_name.endswith('.m4a'):
                    files_info.append({'name': file_name, 'size': file_attr.st_size})
            sftp.close()
            transport.close()
        except Exception as e:
            log_message(f"Error accessing SFTP: {e}")

    # Tworzenie odcinków podcastu
    for file_info in files_info:
        file_name = file_info['name']

        # Wyciąganie daty z nazwy pliku
        try:
            date_part = file_name.split('(')[-1].split(')')[0]
            date = datetime.strptime(date_part, '%d.%m.%y %H.%M')
        except (ValueError, IndexError):
            log_message(f"Failed to parse date from filename: {file_name}")
            date = datetime.now()  # Domyślna data w przypadku błędu

        pub_date = date.strftime('%a, %d %b %Y %H:%M:%S GMT')

        # Dodanie odcinka
        episode = Episode()
        episode.title = f"{show_name} ({date.strftime('%d.%m.%Y %H:%M')})"
        episode.media = Media(
            url=f"{web_base_url.replace(' ', '%20')}/{normalize_filename(station).replace(' ', '%20')}/{normalize_filename(show_name).replace(' ', '%20')}/{file_info['name'].replace(' ', '%20')}",
            size=file_info['size'],
            type="audio/mpeg"
        )
        episode.publication_date = pub_date
        episode.summary = f"Nagranie audycji {show_name} z {station}. Data nagrania: {date.strftime('%d.%m.%Y %H:%M')}"
        podcast.episodes.append(episode)

    if not podcast.episodes:
        log_message("No episodes found. RSS feed will not be generated.")
        return

    # Generowanie pliku RSS
    rss_content = podcast.rss_str()
    local_rss_file = 'feed.xml'
    with open(local_rss_file, 'wb') as f:
        f.write(rss_content.encode('utf-8'))

    # Przesyłanie RSS do destynacji
    try:
        if destination == 'ftp':
            with FTP(target_config['hostname']) as ftp:
                ftp.login(target_config['username'], target_config['password'])
                ftp.cwd(remote_dir)
                with open(local_rss_file, 'rb') as f:
                    ftp.storbinary(f'STOR feed.xml', f)
        elif destination == 'sftp':
            transport = paramiko.Transport((target_config['hostname'].split(':')[0], int(target_config['hostname'].split(':')[1])))
            transport.connect(username=target_config['username'], password=target_config['password'])
            sftp = paramiko.SFTPClient.from_transport(transport)
            sftp.put(local_rss_file, os.path.join(remote_dir, 'feed.xml'))
            sftp.close()
            transport.close()
    except Exception as e:
        log_message(f"Error uploading RSS to {destination.upper()}: {e}")

    # Usuwanie lokalnego pliku RSS
    os.remove(local_rss_file)
    log_message(f"Uploaded RSS file to {destination.upper()}")


def parse_date_from_filename(filename):
    date_str = filename.split('(')[-1].split(')')[0]
    return datetime.strptime(date_str, '%d.%m.%y %H.%M')

def record_show(total_duration, station_url, station, show_name, config, segments=None):
    """
    Nagrywa strumień audio do pliku .ts, nadając mu nazwę zgodną z formatem:
    Test2 (27.12.24 14.49).ts
    """
    recordings_path = config['local']['path']
    dir_path = os.path.abspath(recordings_path) if recordings_path != './' else os.path.dirname(os.path.abspath(__file__))

    date_str = datetime.now().strftime("%d.%m.%y %H.%M")
    output_dir = os.path.join(dir_path, "recordings", normalize_filename(station), normalize_filename(show_name))
    os.makedirs(output_dir, exist_ok=True)
    temp_files = []

    if segments:
        segments_in_seconds = [x * 60 for x in segments]
    else:
        segments_in_seconds = [total_duration * 60]

    for index in range(0, len(segments_in_seconds), 2):
        temp_filename = os.path.join(output_dir, f"{normalize_filename(show_name)}_{index}.ts")
        temp_files.append(temp_filename)

        record_duration = segments_in_seconds[index]
        ffmpeg_command = [
            "ffmpeg",
            "-t", str(record_duration),
            "-reconnect", "1",
            "-reconnect_streamed", "1",
            "-reconnect_at_eof", "1",
            "-reconnect_delay_max", "2",
            "-i", station_url,
            "-c", "copy",
            "-f", "mpegts",
            temp_filename
        ]
        subprocess.run(ffmpeg_command, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        log_message(f"Recorded segment of {segments_in_seconds[index] // 60} minutes: {temp_filename}")

        if index + 1 < len(segments_in_seconds):
            log_message(f"Pause for {segments_in_seconds[index + 1] // 60} minutes.")
            time.sleep(segments_in_seconds[index + 1])

    # Nowa nazwa pliku wynikowego .ts
    ts_filename = f"{normalize_filename(show_name)} ({date_str}).ts"
    combined_ts_file = os.path.join(output_dir, ts_filename)

    with open(os.path.join(output_dir, 'filelist.txt'), 'w') as f:
        for temp_file in temp_files:
            f.write(f"file '{temp_file}'\n")

    ffmpeg_concat_command = [
        "ffmpeg",
        "-f", "concat",
        "-safe", "0",
        "-i", os.path.join(output_dir, 'filelist.txt'),
        "-c", "copy",
        combined_ts_file
    ]
    subprocess.run(ffmpeg_concat_command, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

    log_message(f"Concatenated TS files: {combined_ts_file}")

    for temp_file in temp_files:
        os.remove(temp_file)
    os.remove(os.path.join(output_dir, 'filelist.txt'))

    return combined_ts_file

def main():
    parser = ArgumentParser()
    parser.add_argument("station", type=str, help="Station name")
    parser.add_argument("show_name", type=str, help="Name of the show")
    parser.add_argument("total_duration", type=int, help="Total duration of the recording in minutes")
    parser.add_argument("-p", "--pauses", type=str, help="Sequence of recording and pause times, e.g., '10,2,8,1'")
    parser.add_argument("-s", "--save", nargs='+', choices=['ftp', 'nextcloud', 'local', 'sftp'], help="Save destinations", default=[])
    args = parser.parse_args()

    config = load_config()
    stations = load_stations()
    station_url = stations[args.station]

    if args.pauses:
        segments = list(map(int, args.pauses.split(',')))
    else:
        segments = [args.total_duration]

    combined_ts_file = record_show(args.total_duration, station_url, args.station, args.show_name, config, segments)

    conversion_map = group_destinations_by_params(config, args.save)
    converted_files = convert_audio_once(combined_ts_file, config, conversion_map)
    process_destinations(converted_files, config, args.station, args.show_name, conversion_map)

    os.remove(combined_ts_file)
    log_message("Recording and uploads complete.")

if __name__ == "__main__":
    main()
